@@ -1,9 +1,10 @@
 
 from mistletoe.html_renderer import HTMLRenderer
+# from mistletoe.block_token import BlockCode
 # from subprocess import Popen, PIPE
 # import select
 import base64
-from shutil import copyfile
+from shutil import move
 import rich
 import re
 import os
@@ -19,6 +20,7 @@ import yaml
 from pygments import highlight
 from pygments.lexers import get_lexer_by_name
 from pygments.formatters import HtmlFormatter
+from pygments.util import ClassNotFound
 
 
 FORMAT = "%(message)s"
@@ -42,10 +44,10 @@ pre {
 
 codetemplate = '<pre class="languag-{lang}"><code class="language-{lang}">{inner}</code></pre>'
 divtemplate = '<div class="output" >' + codetemplate + '</div>'
-imgtemplate = '<img src="{src}" class="{ext}  {ucls}"/>'
+# imgtemplate = '<img src="{src}" class="{ext}  {ucls}"/>'
 
-# autocanvast = 'TCanvas *canvas = new TCanvas( "c", "c", 1200, 900 )'
-# autodrawt = 'if (canvas) { canvas->cd(0); canvas->Print("auto.svg"); }\n'
+autocanvast = ''
+autodrawt = 'if (gPad) { gPad->Print("auto.svg"); }\n'
 
 
 
@@ -98,21 +100,29 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
         if self.asset_dir != "" and not self.embed:
             print( "cp %s %s" % (path, self.asset_dir) )
             try :
-                copyfile( path, os.path.join( self.asset_dir, path) )
+                move( path, os.path.join( self.asset_dir, path) )
             except Exception as e:
-                log.error( 'Tried copyfile( "%s", "%s" ) ' % (path, self.asset_dir) )
+                log.error( 'Tried move( "%s", "%s" ) ' % (path, self.asset_dir) )
                 log.error( e )
 
         template = '<img src="data:image/{ext};charset=utf-8;base64,{data}" class="{cls} {ucls}"/>'
-        if self.embed:
+        # cml line flag, block option, or YAML
+        if self.embed or kwargs.get( "embed", False ):
+            # SVG does not need to be b64 embeded in an img tag, just load into DOM 
+            if "svg" == ext:
+                with open(path, "rb") as image_file:
+                    return ''.join(image_file.read().decode('utf-8'))
+
+            # b64 encode other image formats
             with open(path, "rb") as image_file:
                 b64_encoded = base64.b64encode(image_file.read())
-                template = '<img src="data:image/{ext};charset=utf-8;base64,{data}" class="{cls} {ucls}"/>'
-
-                if "svg" == ext:
-                    ext = "svg+xml"
-
-                return template.format( ext=ext, data=b64_encoded.decode(), cls=ext, ucls=kwargs.get("ucls", "") )
+            template = '<img src="data:image/{ext};charset=utf-8;base64,{data}" class="{cls} {ucls}"/>'
+            if self.clean:
+                log.info( "Removing embedded asset: %s" % path )
+                os.remove( path )
+            if "svg" == ext:
+                ext = "svg+xml"
+            return template.format( ext=ext, data=b64_encoded.decode(), cls=ext, ucls=kwargs.get("ucls", "") )
         
 
         if self.clean:
@@ -120,19 +130,22 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
                 os.remove( path )
             except Exception as e:
                 rich.inspect(e)
-
-        return "\n" + imgtemplate.format( src=path, ext=ext, ucls=kwargs.get("ucls", "") )
+        # Return a link to standard image formats
+        image_exts = [ "png", "jpeg", "jpg", "svg", "gif", "webp", "apng" ]
+        if ext in image_exts :
+            template = '<img src="{src}" class="{ext}  {ucls}"/>'
+            return "\n" + template.format( src=path, ext=ext, ucls=kwargs.get("ucls", "") )
+        
+        # try to embed as an object (e.g. PDF)
+        template = '<object data="{src}" class="{ext}" {attr}> </object>'
+        return template.format( src=path, ext=ext, attr="" )
     
     def divWrap(self, inner, cls="", id=""):
-        output = "<div "
-        if "" != cls:
-            output = output + 'class="%s" ' % cls
-        if "" != id:
-            output = output + 'id="%s" ' % id
-        output = output + ">\n"
-        output = output + inner + "\n"
-        output = output + "</div>"
-        return output
+        template = '<div {i} class="{c}" >\n{inner}\n</div>'
+        # avoid empty id attr
+        if id != "":
+            id = 'id="' + id + '"'
+        return template.format( i=id, c=cls, inner=inner )
 
     def render_yaml_fence(self, token):
         log.info("YAML Fence")
@@ -148,12 +161,29 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
             return ""
         return super().render_block_code(token)
 
-    def hpygments( self, token ):
-        code = token.children[0].content
-        lexer = get_lexer_by_name( token.language)
-        formatter = HtmlFormatter( style=self.yaml.get( "pygments-theme", "dracula" ))
+    def pycode( self, code, lang, options ):
+        log.info( "Pygments with theme: %s" % self.yaml.get( "pygments-theme", "dracula" ) )
+        lexer = get_lexer_by_name( lang)
+        try:
+            ln = False
+            if options.get("linenos", False) == "table":
+                ln = "table"
+            elif options.get("linenos", False) != False:
+                ln = "inline"
+
+            formatter = HtmlFormatter( style=self.yaml.get( "pygments-theme", "dracula" ), linenos=ln, wrapcode=True )
+            return highlight(code, lexer, formatter)
+        except ClassNotFound as e:
+            log.warning( "Cannot find pygment style: %s " % self.yaml.get( "pygments-theme", "dracula" ) )
+            log.info( "Using default style as fallback" )
+            from pygments.styles import STYLE_MAP
+            log.info( "Available styles are:" )
+            log.info( STYLE_MAP )
+        formatter = HtmlFormatter( style="default")
         return highlight(code, lexer, formatter)
-        
+
+    def hpygments( self, token ):
+        return self.pycode( token.children[0].content, token.language, token.options)
 
     def process_yaml( self, yaml ):
         self.yaml = yaml
@@ -162,11 +192,33 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
         self.description = yaml.get( "description", self.description )
         self.keywords = yaml.get( "keywords", self.keywords )
         self.autodraw = yaml.get( "auto-draw", self.autodraw )
+        self.highlighter = self.yaml.get("highlight", "prismjs").lower()
 
+        if self.autodraw:
+            self.embed = True
 
-        # if self.autodraw:
-        #     self.run_cmd( autocanvast )
+    def render_inline_code( self, token):
+        # rich.inspect( token )
+        code = token.children[0].content
 
+        # evaluate code inside BASH style blocks
+        m = re.search( "\${(.*?)}", code )
+        if m:
+            # log.info( "--->CODE: %s" % ( m.groups()[0] ) )
+            evalcode = m.groups()[0]
+            evalcode += ';printf("\\n");'
+
+            log.info( 'eval: "%s"' % evalcode )
+            output, err, imgs = self.run_cmd( evalcode )
+            log.info( "stdout: %s" % ( output ) )
+            log.info( "stderr: %s" % ( err ) )
+            return output
+
+        return super().render_inline_code( token )
+
+    """
+    Process gnuPlot code execution
+    """
     def process_gnuplot( self, token ):
         log.info( "gnuplot" )
         token.language = "python" # just for output syntax highlighting
@@ -182,13 +234,20 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
         self.gnuplot.run( code )
 
         # inject images
-        imgout = '<div id="{id}" class="root-images" style="text-align: center;">'.format( id="root-images-%d" % (self.blockid))
+        imgout = '<div id="{id}" class="image-out gnuplot-images" style="text-align: center;">'.format( id="gnuplot-images-%d" % (self.blockid))
         for i in imgs:
             imgout += self.process_image_output( i, ucls=img_class )
         imgout += '</div>'
 
-        return code_block + self.divWrap(imgout, "gnuplot-block", "gnuplot-output-block-%d" % (self.blockid - 1) )
+        return self.divWrap( code_block, "gnuplot-block input-block", "" ) + self.divWrap(
+                imgout, 
+                "gnuplot-block output-block", 
+                "gnuplot-output-block-%d" % (self.blockid) 
+            )
 
+    """
+    Process RNUPlot code execution
+    """
     def process_rnuplot( self, token ):
         log.info( "rnuplot" )
         token.language = "python" # just for output syntax highlighting
@@ -209,8 +268,7 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
             imgout += self.process_image_output( i, ucls=img_class )
         imgout += '</div>'
 
-        return code_block + self.divWrap(imgout, "gnuplot-block", "gnuplot-output-block-%d" % (self.blockid - 1) )
-    
+        return code_block + self.divWrap(imgout, "gnuplot-block", "gnuplot-output-block-%d" % (self.blockid) )
 
     """
     Code Fence is a custom Code Block token that accepts optional arguments after the language id
@@ -223,13 +281,16 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
     """
     def render_code_fence( self, token ):
         log.info( "render_code_fence" )
+        # rich.inspect( token )
         return self.render_block_code( token )
 
-    
     def optionAsBool( self, options, n, default = False ):
+        
         # option is false by default
         if n not in options :
             return default
+        if type( options[n] ) == bool:
+            return options[n]
         return  not (options.get( n, "" ).lower() == 'false' or options.get( n, "" ) == '0')
 
     def process_cpp( self, token ):
@@ -239,20 +300,25 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
             code_block =  super().render_block_code(token)
         code = token.children[0].content
         
-        log.info( "exec: %r" % self.optionAsBool( token.options, "exec", True ) )
-        log.info ("no-exec: %r" % self.no_exec)
+        # log.info( "exec: %r" % self.optionAsBool( token.options, "exec", True ) )
+        # log.info ("no-exec: %r" % self.no_exec)
         if self.optionAsBool( token.options, "exec", True ) == False or self.no_exec:
             log.info( "skipping execution of cpp code block" )
             return code_block
 
         # rich.inspect( token.options )
         # optional class for input code block
-        input_class = token.options.get( ".in", "" ) + " root-block root-block-green"
+        input_class = token.options.get( ".in", "" ) + " root-block input-block"
         code_block = self.divWrap( code_block, input_class)
 
+        if self.autodraw:
+            code = autocanvast + code + autodrawt
+
         # Execute the codez!
         output, err, imgs = self.run_cmd( code )
-        output = ("# Block [%d]\n" % self.blockid) + output
+        raw_output = output
+        if len(output):
+            output = ("# Block [%d]\n" % self.blockid) + output
 
         # output controls
         if self.optionAsBool( token.options, "out", True) == False or self.optionAsBool( token.options, "stdout", True) == False:
@@ -270,7 +336,9 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
             with open( stderr_filename, "w" ) as wf:
                 wf.writelines( err.split() )
         
-        log.info( "silent? %r" % ('silent' in token.options) )
+        log.error( err )
+
+        # log.info( "silent? %r" % ('silent' in token.options) )
         if 'silent' in token.options or 'quiet' in token.options:
             output = ""
             err = ""
@@ -279,30 +347,51 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
             code_block = ""
         
         img_class = token.options.get( ".image", "" )
-        if self.optionAsBool( token.options, 'img', True ) == False:
+        if self.optionAsBool( token.options, 'img', self.optionAsBool( token.options, 'image', True ) ) == False:
             imgs = []
 
-        # inject stdoutput 
-        divout = '<div id="{id}" class="root-output" style="text-align: center;">'.format( id="root-output-%d" % (self.blockid) )
-        if len( output + err ):
-            divout += "\n" + divtemplate.format( lang="sh", inner=self.escape_html(output + err) )
+        # inject stdout
+        divout = '<div id="{id}" class="root-output output-block">'.format( id="root-output-%d" % (self.blockid) )
+        if len( output ):
+            if self.highlighter == "pygments":
+                divout += self.pycode( output, "bash", {} )
+            else:
+                divout +=  "\n" + divtemplate.format( lang="bash", inner=self.escape_html(output) )
+        divout += '</div>'
+
+        divout += '<div id="{id}" class="root-output output-block-err">'.format( id="root-err-%d" % (self.blockid) )
+        if len( err ):
+            if self.highlighter == "pygments":
+                divout += self.pycode( err, "bash", {} )
+            else:
+                divout += "\n" + divtemplate.format( lang="bash", inner=self.escape_html(err) )
         divout += '</div>'
 
         # inject images
-        imgout = '<div id="{id}" class="root-images" style="text-align: center;">'.format( id="root-images-%d" % (self.blockid))
+        imgout = '<div id="{id}" class="root-images output-images">'.format( id="root-images-%d" % (self.blockid))
         for i in imgs:
             imgout += self.process_image_output( i, ucls=img_class )
         imgout += '</div>'
 
-        lexer = get_lexer_by_name("c++")
-        formatter = HtmlFormatter( style=self.yaml.get( "pygments-theme", "dracula" ))
-        if code_block != ""  and self.yaml.get("highlight", "prismjs").lower() == "pygments" :
-            code_block = highlight(code, lexer, formatter)
+
+        # lexer = get_lexer_by_name("c++")
+        # formatter = HtmlFormatter( style=self.yaml.get( "pygments-theme", "dracula" ))
+        # if code_block != ""  and self.yaml.get("highlight", "prismjs").lower() == "pygments" :
+        #     code_block = highlight(code, lexer, formatter)
 
         self.blockid = self.blockid + 1
+
+        if token.options.get( "raw" ) == True:
+            log.info( "RAW OUTPUT: %s" % raw_output )
+            return raw_output
+
         return code_block + self.divWrap( divout + imgout, "root-block", "root-output-block-%d" % (self.blockid - 1) )
 
     def process_js(self, token):
+        if self.highlighter == "pygments":
+            code_block = self.hpygments( token )
+        else:
+            code_block =  super().render_block_code(token)
         code =token.children[0].content
         template = '<script>\n{content}\n</script>'
         if "//qin" in code:
@@ -313,7 +402,11 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
         return code_block + template.format(content=code)
 
     def process_css(self, token):
-        code =token.children[0].content
+        if self.highlighter == "pygments":
+            code_block = self.hpygments( token )
+        else:
+            code_block =  super().render_block_code(token)
+        code = token.children[0].content
         template = '<style>\n{content}\n</style>'
         if "/* qin */" in code or "/*qin*/" in code:
             code_block = "" # dont output the code
@@ -323,6 +416,10 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
         return code_block + template.format(content=code)
 
     def process_html(self, token):
+        if self.highlighter == "pygments":
+            code_block = self.hpygments( token )
+        else:
+            code_block =  super().render_block_code(token)
         code =token.children[0].content
         template = '<div>\n{content}\n</div>'
         if "<!--qin-->" in code or "<!-- qin -->" in code:
@@ -335,16 +432,17 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
 
     def render_block_code(self, token):
         log.info( 'block_code' )
-        
-        # code =token.children[0].content
-        # if self.autodraw :
-        #     code += autodrawt
-        # rich.inspect( token )
-        # log.info( code )
-        # if token.language:
-        #     attr = ' class="{}"'.format('language-{}'.format(self.escape_html(token.language)))
-        # else:
-        #     attr = ''
+        # rich.inspect( token.options )
+
+        # include a file into a code block
+        include = token.options.get( "include", "" )
+        if "include" in token.options and os.path.exists( include ):
+            log.info( "including file: %s" % include )
+            with open( include ) as f:
+                c = f.read()
+                token.children[0].content = c
+        elif "include" in token.options and not os.path.exists( include ):
+            log.warning( "Cannot include %s, file DNE" % include )
 
         if "gnuplot" == token.language or token.options.get( "exec", "" ) == "gnuplot":
             return self.process_gnuplot( token )
@@ -364,7 +462,10 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
         if "html" == token.language:
             return self.process_html(token)
 
-        code_block =  super().render_block_code(token)
+        if self.highlighter == "pygments":
+            code_block = self.hpygments( token )
+        else:
+            code_block =  super().render_block_code(token)
         return code_block
     
     def loadCSS( self ):
@@ -396,7 +497,10 @@ class RootHtmlRenderer(HTMLRenderer, RootExecutor):
                 log.error( e )
         
         if self.yaml.get("highlight", "prismjs").lower() == "pygments":
-            css = css + HtmlFormatter(style=self.yaml.get( "pygments-theme", "dracula" )).get_style_defs('.highlight') + pygments_extra_css
+            try :
+                css = css + HtmlFormatter(style=self.yaml.get( "pygments-theme", "dracula" )).get_style_defs('.highlight') + pygments_extra_css
+            except ClassNotFound as e:
+                log.warning( "Cannot find pygments style %s" % self.yaml.get( "pygments-theme", "dracula" ) )
         html = ""
         try :
             with open( os.path.join(mydir, "data/template.html")) as f:
