@@ -139,11 +139,16 @@ class CodeFence(BlockToken):
     pattern = re.compile(r'( {0,3})(`{3,}|~{3,}) *(\S*)')
     _open_info = None
     _options = {}
-    def __init__(self, match):
-        lines, open_info, options = match
-        self.language = span_token.EscapeSequence.strip(open_info[2])
-        self.children = (span_token.RawText(''.join(lines)),)
-        self.options = options
+    def __init__(self, match=None):
+        if match == None:
+            self.language = ""
+            self.children = []
+            self.options = {}
+        else:
+            lines, open_info, options = match
+            self.language = span_token.EscapeSequence.strip(open_info[2])
+            self.children = (span_token.RawText(''.join(lines)),)
+            self.options = options
 
     @classmethod
     def start(cls, line):
